#if !defined(LKT_COMMON_H__)
#define LKT_COMMON_H__

#include <lektor/internal/os.h>

#if defined(__cplusplus)
#define LKT_INCLUDE_CXX
extern "C" {
#endif

/* Hide the DB implementation */
#include <sqlite3.h>
typedef volatile sqlite3 lkt_db;

#define NOT_IMPLEMENTED ___not_implemented(__func__, __FILE__, __LINE__)
extern EXIT_FUNCTION ___not_implemented(const char *func, char *file, int line);

#define SET_FLAG(i, flag)   (i) |= (flag)
#define UNSET_FLAG(i, flag) (i) &= ~(flag)
#define HAS_FLAG(i, flag)   ((bool)((i) & (flag)))

/* TODO macro, use like:
 * #pragma message(TODO "What to do")*/
#define _QUOTE(x) #x
#define QUOTE(x)  _QUOTE(x)
#define TODO      __FILE__ "(+" QUOTE(__LINE__) ") : "

/* Max value for any buffer, to not squash the stack. */
#define BUFFER_MAX 4096

#ifndef MAX
#define MAX(a, b) ((a) < (b) ? (b) : (a))
#endif /* MAX */

#ifndef MIN
#define MIN(a, b) ((a) > (b) ? (b) : (a))
#endif /* MIN */

#if defined(LKT_INCLUDE_CXX)
#define ___LKT_TO_BOOL(cond) static_cast<bool>(static_cast<int>(cond))
#else
#define ___LKT_TO_BOOL(cond) (bool)(int)(cond)
#endif

#define RETURN_IF(cond, msg, ret)      \
    if (___LKT_TO_BOOL(cond)) {        \
        LOG_DEBUG("DEBUG", "%s", msg); \
        return ret;                    \
    }
#define GOTO_IF(cond, msg, label)      \
    if (___LKT_TO_BOOL(cond)) {        \
        LOG_DEBUG("DEBUG", "%s", msg); \
        goto label;                    \
    }
#define FAIL_IF(cond, ...)          \
    {                               \
        if (___LKT_TO_BOOL(cond)) { \
            LOG_FATAL(__VA_ARGS__); \
        }                           \
    }

#define unless(cond) if (!(cond))

#define FAIL_UNLESS(cond, ...)        FAIL_IF(!(cond), __VA_ARGS__)
#define GOTO_UNLESS(cond, msg, label) GOTO_IF(!(cond), msg, label)
#define RETURN_UNLESS(cond, msg, ret) RETURN_IF(!(cond), msg, ret)
#define NOTHING                       /* Usefull to return nothing in previous macros */

#define DO_WITH_LOCK(lock, error, instruction)                             \
    {                                                                      \
        GOTO_IF(pthread_mutex_lock(&(lock)), "Failed to lock", error);     \
        instruction;                                                       \
        GOTO_IF(pthread_mutex_unlock(&(lock)), "Failed to unlock", error); \
    }

#define DO_WITH_RELAXED_LOCK(lock, error, instruction)                     \
    {                                                                      \
        GOTO_IF(pthread_mutex_unlock(&(lock)), "Failed to unlock", error); \
        instruction;                                                       \
        GOTO_IF(pthread_mutex_lock(&(lock)), "Failed to lock", error);     \
    }

#define STRTOL_ERROR(integer) (errno == ERANGE || (errno != 0 && (integer) == 0))
#define STRTOL(ret, str, endptr, err_flag)                            \
    {                                                                 \
        (err_flag) = 0;                                               \
        errno      = 0;                                               \
        (ret)      = ((str) == NULL) ? 0 : strtol(str, &(endptr), 0); \
        err_flag   = errno != 0 || endptr == str;                     \
    }

/* Get a generic function pointer from any function pointer. */
#define FUNCTION_POINTER(func) ((void (*)(void))(func))

/* Forever macros */
#define FOR_EVER for (;;)
#define FOR_EVER_IF(expr) \
    FOR_EVER              \
    {                     \
        if (!(expr))      \
            break;        \
    }
#define FOR_EVER_UNTIL(expr) FOR_EVER_IF (!(expr))

/* For each, for flat-list. Need to be null terminated on the normally non-null field. */
#define FOR_EACH_FLAT_LIST_ITEM(it, nnull_field) for (; (it)->nnull_field != NULL; it++)

/* For each argument in a thing like a `char *args[]` */
#define FOR_EACH_ARGUMENT(i, args) for (int i = 0; args[i]; ++i)

/* Custom defined assert. */
extern void (*___lkt_assert)(const char *file, int line, const char *func, const char *msg);
#ifdef assert
#undef assert
#endif
#define lkt_assert(ex) (void)((ex) || (___lkt_assert(__FILE__, __LINE__, __func__, #ex), 0))
#ifdef NDEBUG
#define assert(ex)
#else
#define assert(ex) lkt_assert(ex)
#endif

/* Custom defined abort. */
extern EXIT_FUNCTION ___lkt_abort(void);
#ifdef abort
#undef abort
#endif
#define abort() ___lkt_abort()

/* Custom log functions. */
typedef enum {
    LOG_LEVEL_ERROR = 1,
    LOG_LEVEL_WARN  = 2,
    LOG_LEVEL_INFO  = 3,
    LOG_LEVEL_DEBUG = 4,
    ___LAST_UNUSED_LOG_LEVEL, /* Usefull to enable all by default */
} LOG_LEVEL;
struct lkt_logfile;

void lkt_set_log_level_verbose(bool);
void lkt_set_log_level(LOG_LEVEL);
void lkt_set_log_logfile(struct lkt_logfile *);
LOG_LEVEL lkt_get_log_level(void);

void ___lkt_log(LOG_LEVEL, const char *sec, const char *func, const char *file, uint64_t l,
                const char *fmt, ...);
#define LOG_DEBUG(sec, ...) \
    ___lkt_log(LOG_LEVEL_DEBUG, sec, __func__, __FILE__, __LINE__, __VA_ARGS__)
#define LOG_INFO(sec, ...) \
    ___lkt_log(LOG_LEVEL_INFO, sec, __func__, __FILE__, __LINE__, __VA_ARGS__)
#define LOG_WARN(sec, ...) \
    ___lkt_log(LOG_LEVEL_WARN, sec, __func__, __FILE__, __LINE__, __VA_ARGS__)
#define LOG_ERROR(sec, ...) \
    ___lkt_log(LOG_LEVEL_ERROR, sec, __func__, __FILE__, __LINE__, __VA_ARGS__)
#define LOG_FATAL(...)                   \
    {                                    \
        LOG_ERROR("FATAL", __VA_ARGS__); \
        exit(1);                         \
    }
#define LOG_UNREACHABLE LOG_FATAL("Ureachable!");
#define LOG_DEBUG_IF(cond, sec, ...)     \
    {                                    \
        if (cond)                        \
            LOG_DEBUG(sec, __VA_ARGS__); \
    }
#define LOG_INFO_IF(cond, sec, ...)     \
    {                                   \
        if (cond)                       \
            LOG_INFO(sec, __VA_ARGS__); \
    }
#define LOG_WARN_IF(cond, sec, ...)     \
    {                                   \
        if (cond)                       \
            LOG_WARN(sec, __VA_ARGS__); \
    }
#define LOG_ERROR_IF(cond, sec, ...)     \
    {                                    \
        if (cond)                        \
            LOG_ERROR(sec, __VA_ARGS__); \
    }

void ___lkt_print(const char *section, const char *format, ...);
#define LKT_OUTPUT(section, ...) ___lkt_print(section, __VA_ARGS__)

#define LKT_ENV_RESTART      "__LKT_RESTART"
#define LKT_ENV_CURRENT      "__LKT_CURRENT"
#define LKT_ENV_UNSET_ASSERT "__LKT_UNSET_ASSERT"
#define LKT_ENV_KLKT_PID     "__LKT_KLKT_PID"

#define LKT_TABLE_FIELDS_MAX       20
#define LKT_LINE_MAX               1024
#define LKT_MAX_SQLITE_STATEMENT   1024
#define LKT_PROTECTED_DATABASE     "disk"
#define LKT_PROTECTED_DATABASE_NEW "newdisk"

#define LKT_DB_ID     "id"
#define LKT_DB_NAME   "source_name"
#define LKT_DB_TITLE  "song_title"
#define LKT_DB_CAT    "category"
#define LKT_DB_TYPE   "song_type"
#define LKT_DB_AUTHOR "author_name"
#define LKT_DB_LANG   "language"
#define LKT_DB_ALL    "string"

/* Default iteration count in search_iter */
#define LKT_DB_DEFAULT_ITERATION_MAX 100

#define LEKTOR_TAG_MAX        256  /* Max length of a tag                          */
#define LKT_COMMAND_LEN_MAX   32   /* MPD commands are at most 32 character long   */
#define LKT_MESSAGE_ARGS_MAX  32   /* At most 32 words in a command are supported  */
#define LKT_MESSAGE_MAX       2048 /* A message is at most <defined> chars long    */
#define LKT_DEFAULT_LIST_SIZE 10

#define LKT_BACKLOG      32
#define COMMAND_LIST_MAX 64 /* At most 64 commands per client */
#define BUFFER_OUT_MAX   16 /* At most 16 messages per client */
#define MPD_VERSION      "0.21.16"

typedef enum {
    MPD_IDLE_NONE = 0,

    MPD_IDLE_DATABASE        = (1 << 1),
    MPD_IDLE_UPDATE          = (1 << 2),
    MPD_IDLE_STORED_PLAYLIST = (1 << 3),
    MPD_IDLE_PLAYLIST        = (1 << 4),
    MPD_IDLE_PLAYER          = (1 << 5),
    MPD_IDLE_MIXER           = (1 << 6),
    MPD_IDLE_OUTPUT          = (1 << 7),
    MPD_IDLE_OPTIONS         = (1 << 8),
    MPD_IDLE_PARTITION       = (1 << 9),
    MPD_IDLE_STICKER         = (1 << 10),
    MPD_IDLE_SUBSCRIPTION    = (1 << 11),
    MPD_IDLE_MESSAGE         = (1 << 12),

    MPD_IDLE_ALL = MPD_IDLE_DATABASE | MPD_IDLE_UPDATE | MPD_IDLE_STORED_PLAYLIST |
                   MPD_IDLE_PLAYLIST | MPD_IDLE_PLAYER | MPD_IDLE_MIXER | MPD_IDLE_OUTPUT |
                   MPD_IDLE_OPTIONS | MPD_IDLE_PARTITION | MPD_IDLE_STICKER |
                   MPD_IDLE_SUBSCRIPTION | MPD_IDLE_MESSAGE,
} MPD_IDLE_FLAG;

#define STR_MATCH(str1, str2)     (!strcasecmp(str1, str2))
#define STR_NMATCH(str1, str2, n) (!strncasecmp(str1, str2, n))

/* Get the number of args in __VA_ARGS__ */
#define GET_VA_COUNT(...) PP_NARG_(__VA_ARGS__, PP_RSEQ_N())
#define PP_NARG_(...)     PP_ARG_N(__VA_ARGS__)
#define PP_ARG_N(_1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12, _13, _14, _15, _16, _17, _18, \
                 _19, _20, _21, _22, _23, _24, _25, _26, _27, _28, _29, _30, _31, _32, _33, _34,  \
                 _35, _36, _37, _38, _39, _40, _41, _42, _43, _44, _45, _46, _47, _48, _49, _50,  \
                 _51, _52, _53, _54, _55, _56, _57, _58, _59, _60, _61, _62, _63, N, ...)         \
    N
#define PP_RSEQ_N()                                                                             \
    63, 62, 61, 60, 59, 58, 57, 56, 55, 54, 53, 52, 51, 50, 49, 48, 47, 46, 45, 44, 43, 42, 41, \
        40, 39, 38, 37, 36, 35, 34, 33, 32, 31, 30, 29, 28, 27, 26, 25, 24, 23, 22, 21, 20, 19, \
        18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0

int is_utf8(const char *string);

/* Trim the string `str` of the char `c` from the right and the left.
   The string may not be null terminated.
   Returns a pointer to the the trimmed string
   Restrictions: the len of the string must be `len`. */
char *trim(char *str, size_t len, char c);

long long_length(long integer);

/* Safe functions, which preserve the terminal '\0'. */
void *safe_malloc(size_t size);
void *safe_realloc(void *ptr, unsigned long size);
void safe_free(void **);
int safe_snprintf(char *dest, const size_t max_len, const char *format, ...);
int safe_vsnprintf(char *dest, const size_t max_len, const char *format, va_list);
char *safe_strncpy(char *dest, const char *src, size_t n);

/* Helpers for malloc functions. */
#if defined(LKT_INCLUDE_CXX)
#define LKT_ALLOC(___struct___)        static_cast<___struct___ *>(safe_malloc(sizeof(___struct___)))
#define LKT_ALLOC_STRUCT(___struct___) LKT_ALLOC(struct ___struct___)
#define LKT_ALLOC_ARRAY(___struct___, ___count___) \
    static_cast<___struct___ *>(safe_malloc(sizeof(___struct___) * (size_t)(___count___)))
#define LKT_ALLOC_STRUCT_ARRAY(___struct___, ___count___) \
    LKT_ALLOC_ARRAY(struct ___struct___, ___count___)
#else
#define LKT_ALLOC(___struct___)        (___struct___ *)safe_malloc(sizeof(___struct___))
#define LKT_ALLOC_STRUCT(___struct___) LKT_ALLOC(struct ___struct___)
#define LKT_ALLOC_ARRAY(___struct___, ___count___) \
    (___struct___ *)safe_malloc(sizeof(___struct___) * (size_t)(___count___))
#define LKT_ALLOC_STRUCT_ARRAY(___struct___, ___count___) \
    LKT_ALLOC_ARRAY(struct ___struct___, ___count___)
#endif

/* Iterate throught a string, copy each element in the dest string.
   The save ptr must contains a null integer. */
int iter_string(const char *str, const char *sep, char *dest, const size_t dest_len, size_t *save);

#if defined(__cplusplus)
}
#endif
#endif /* LKT_COMMON_H__ */
