#include <lektor/logfile.h>

struct lkt_logfile {
    const char *base_folder; /* Allocated, must be freed */
    const unsigned int file_count;
    const unsigned int line_count;
    int current_fd;                     /* The <folder/1> file */
    unsigned int current_lines_in_file; /* Size is file_count */
};

struct lkt_logfile *___log_logfile = NULL; /* Global logfile, use if necessary */

DESTRUCTOR_FUNCTION
___close_global_logfile(void)
{
    if (___log_logfile) {
        struct lkt_logfile *logfile = ___log_logfile;
        ___log_logfile              = NULL;
        lkt_logfile_free(logfile);
    }
}

PRIVATE_FUNCTION unsigned int
___count_lines_in_file(const char *file)
{
    unsigned int lines = 0u;
    FILE *fd           = fopen(file, "r");

    /* Reads should be buffured... */
    while ((fd != NULL) && !feof(fd))
        lines += (fgetc(fd) == '\n');

    return lines;
}

PRIVATE_FUNCTION void
___close_all_files(struct lkt_logfile *logfile)
{
    if (logfile->current_fd > 0) {
        fsync(logfile->current_fd);
        close(logfile->current_fd);
    }
}

PRIVATE_FUNCTION void
___open_all_files(struct lkt_logfile *logfile)
{
    char path[PATH_MAX];
    safe_snprintf(path, PATH_MAX, "%s/1", logfile->base_folder);
    logfile->current_lines_in_file = ___count_lines_in_file(path);
    logfile->current_fd            = open(path, O_APPEND | O_CREAT | O_WRONLY, 0644);
    FAIL_IF(logfile->current_fd <= 0, "Failed to open log file: %s", path);
}

PRIVATE_FUNCTION void
___rotate_logfile(struct lkt_logfile *logfile)
{
    char oldpath[PATH_MAX];
    char newpath[PATH_MAX];
    memset(oldpath, 0, sizeof(oldpath));
    memset(newpath, 0, sizeof(newpath));

    safe_snprintf(oldpath, PATH_MAX, "%s/", logfile->base_folder);
    safe_snprintf(newpath, PATH_MAX, "%s/", logfile->base_folder);

    const size_t base_folder_len = strlen(logfile->base_folder);
    const size_t remaining_place = PATH_MAX - strlen(oldpath) - 1;
    char *const oldpath_num      = &oldpath[base_folder_len + 1];
    char *const newpath_num      = &newpath[base_folder_len + 1];

    ___close_all_files(logfile);

    /* Move i -> i-1 */
    for (size_t i = logfile->file_count; i > 0; --i) {
        safe_snprintf(oldpath_num, remaining_place, "%u", i);
        safe_snprintf(newpath_num, remaining_place, "%u", i - 1);
        rename(newpath, oldpath);
    }

    ___open_all_files(logfile);

    LOG_INFO("LOG", "Log files rotated");
}

struct lkt_logfile *
lkt_logfile_new(lkt_db *db)
{
    char base_folder[PATH_MAX];
    int file_count;
    int line_count;

    DBCONF_GET_INT(db, "log", "file_count", &file_count);
    DBCONF_GET_INT(db, "log", "line_count", &line_count);
    DBCONF_GET_TEXT(db, "log", "folder", base_folder, PATH_MAX);

    if (file_count < 0 || line_count < 0)
        LOG_FATAL("Negative file or line count in setting for logging");

    struct lkt_logfile *ret        = LKT_ALLOC_STRUCT(lkt_logfile);
    struct lkt_logfile tmp_logfile = {
        .file_count            = (unsigned int)file_count,
        .line_count            = (unsigned int)line_count,
        .base_folder           = strdup(base_folder),
        .current_fd            = 0,
        .current_lines_in_file = 0,
    };

    memcpy(ret, &tmp_logfile, sizeof(struct lkt_logfile));

    errno = 0;
    FAIL_IF(mkdir(base_folder, 0755) && (errno != EEXIST), "Failed to create folder: %s",
            base_folder);
    ___open_all_files(ret);

    LOG_INFO("LOG", "Log folder was created at: %s", base_folder);
    return ret;
}

void
lkt_logfile_write(struct lkt_logfile *logfile, const char *line)
{
    if (logfile->current_lines_in_file >= logfile->line_count)
        ___rotate_logfile(logfile);
    write(logfile->current_fd, line, strlen(line));
    ++(logfile->current_lines_in_file);
}

void
lkt_logfile_vfwrite(struct lkt_logfile *logfile, const char *line, va_list ap)
{
    if (logfile->current_lines_in_file >= logfile->line_count)
        ___rotate_logfile(logfile);
    vdprintf(logfile->current_fd, line, ap);
    ++(logfile->current_lines_in_file);
}

void
lkt_logfile_free(struct lkt_logfile *logfile)
{
    free((void *)logfile->base_folder);
    ___close_all_files(logfile);
    free(logfile);
}
